# -*- coding: utf-8 -*-
"""
Created on Fri Oct 02 08:39:31 2019

@author: Pavel Gostev
@email: gostev.pavel@physics.msu.ru

Версия алгоритма расчета статистики фотоотсчетов
на основе numpy.float128
"""

import numpy as np

from ._numpy_core import normalize
from ._dmatrix import d_binomial, d_subbinomial
from ._dmatrix import invd_binomial, invd_subbinomial


def d_matrix(qe: float, N: int, M: int, mtype: str = 'binomial', n_cells: int = 0):
    """
    Method for construction of binomial or subbinomial photodetection matrix
    with size NxM

    Parameters
    ----------
    qe : float
        Photon Detection Efficiency (PDE) of the detector.
    N : int
        Maximum numbers of photons
    M : int
        Maximum number of photocounts
    mtype : {'binomial', 'subbinomial'}, optional
        Type of the detector: ideal is binomial, realistic is subbinomial,
        but in the most of applications one can consider the detector as binomial
        The default is 'binomial'.
    n_cells : TYPE, optional
        Number of photocounting cells in the subbinomial case. The default is 0.

    Raises
    ------
    ValueError
        Wrong method for matrix construction.
        mtype must be binomial or subbinomial.

    Returns
    -------
    numpy.ndarray
        Binomial or subbinomial photodetection matrix of size NxM.

    """

    if mtype == 'binomial':
        return d_binomial(qe, N, M)
    elif mtype == 'subbinomial':
        return d_subbinomial(qe, N, M, n_cells)
    else:
        raise ValueError("""
                         Can't construct detection matrix of type %s
                         mtype must be binomial or subbinomial
                         """ % mtype)


def invd_matrix(qe: float, N: int, M: int, mtype: str = 'binomial', n_cells: int = 0):
    """
    Method for construction of binomial or
    subbinomial inverse photodetection matrix
    with size MxN

    Parameters
    ----------
    qe : float
        Photon Detection Efficiency (PDE) of the detector.
    N : int
        Maximum numbers of photons
    M : int
        Maximum number of photocounts
    mtype : {'binomial', 'subbinomial'}, optional
        Type of the detector: ideal is binomial, realistic is subbinomial,
        but in most applications one can consider the detector as binomial
        The default is 'binomial'.
    n_cells : TYPE, optional
        Number of photocounting cells is subbinomial case. The default is 0.

    Raises
    ------
    ValueError
        Wrong method for matrix construction.
        mtype must be binomial or subbinomial..

    Returns
    -------
    numpy.ndarray
        Binomial or subbinomial inverse photodetection matrix of size MxN.

    """
    if mtype == 'binomial':
        return invd_binomial(qe, N, M)
    elif mtype == 'subbinomial':
        return invd_subbinomial(qe, N, M, n_cells)
    else:
        raise ValueError("""
                         Can't construct inversed detection matrix of type %s
                         mtype must be binomial or subbinomial
                         """ % mtype)


def P2Q(P: np.ndarray, qe: float, M: int = 0, mtype: str = 'binomial', n_cells: int = 0):
    """
    Method for calculation of photocounting statistics
    from photon-number statistics

    Parameters
    ----------
    P : numpy.ndarray
        Photon-number statistics.
    qe : float
        Photon Detection Efficiency (PDE) of the detector.
    M : int, optional
        Maximum number of photocounts. It's undependent from length of P
        The default is 0, and in this case length(Q) = length(P).
    mtype : {'binomial', 'subbinomial'}, optional
        Type of the detector: ideal is binomial, realistic is subbinomial,
        but in most applications one can consider the detector as binomial
        The default is 'binomial'.
    n_cells : TYPE, optional
        Number of photocounting cells is subbinomial case. The default is 0.

    Returns
    -------
    Q : numpy.ndarray
        Photocounting statistics.

    """

    N = len(P)
    if M == 0:
        M = N
    return normalize(d_matrix(qe, N, M, mtype, n_cells).dot(P))


def Q2P(Q: np.ndarray, qe: float, N=0, mtype='binomial', n_cells=0):
    """
    Method for calculation of photon-number statistics
    from photocounting statistics by simple inversion

    Parameters
    ----------
    Q : numpy.ndarray
        Photocounting statistics.
    qe : float
        Photon Detection Efficiency (PDE) of the detector.
    N : int, optional
        Maximum number of photons. It's undependent from length of Q
        The default is 0, and in this case length(Q) = length(P).
    mtype : {'binomial', 'subbinomial'}, optional
        Type of the detector: ideal is binomial, realistic is subbinomial,
        but in most applications one can consider the detector as binomial
        The default is 'binomial'.
    n_cells : TYPE, optional
        Number of photocounting cells is subbinomial case. The default is 0.

    Returns
    -------
    P : numpy.ndarray
        Photon-number statistics.

    """
    M = len(Q)
    if N == 0:
        N = M
    if N > M:
        Q = np.concatenate((Q, np.zeros(N - M)))
        M = N

    return normalize(invd_matrix(qe, M, N, mtype, n_cells).dot(Q))



# -*- coding: utf-8 -*-
"""
Created on Thu Jul  4 16:25:54 2019

@author: Pavel Gostev
"""
from os.path import isfile, join
from os import listdir
import time

import gc

from dataclasses import dataclass

import lecroyparser
import tekwfm2 as tek

from scipy.sparse.linalg import spsolve
from scipy import sparse
from scipy.signal import find_peaks

import numpy as np

from compress_pickle import dump, load
from joblib import Parallel, delayed

gc.enable()

VENDOR_EXT = {'tek': 'wfm',
              'lecroy': 'trc'}


def parse_file(datafile, vendor):
    if vendor == 'lecroy':
        data = lecroyparser.ScopeData(datafile)
        delattr(data, "file")
    if vendor == 'tek':
        data = tek.ScopeData(datafile)
    return data


def list_files(datadir, vendor, fsoffset, fsnum):
    lf = [join(datadir, f) for f in listdir(datadir)
          if isfile(join(datadir, f)) and f.endswith(VENDOR_EXT[vendor])]
    if fsoffset != 0:
        lf = lf[fsoffset:]
    if fsnum > 0:
        lf = lf[:fsnum]
    return lf


def parse_files(oscfiles, vendor, fsnum=0, parallel=False):
    n_jobs = -1 if parallel else 1

    if fsnum > 0:
        oscfiles = oscfiles[:fsnum]
    return Parallel(
        n_jobs=n_jobs)([delayed(parse_file)(df, vendor) for df in oscfiles])


def windowed(data, div_start, div_width):
    div_points = len(data.x) / data.horizInterval
    wstart = int(div_start * div_points)
    wwidth = int(div_width * div_points)
    return data.x[wstart:wstart + wwidth], data.y[wstart:wstart + wwidth]


def baseline_als(y, lam, p, niter=2):
    """
    See "Asymmetric Least Squares Smoothing" by P. Eilers and H. Boelens in 2005
    https://www.researchgate.net/publication/228961729_Baseline_Correction_with_Asymmetric_Least_Squares_Smoothing
    """
    L = len(y)
    D = sparse.diags([1, -2, 1], [0, -1, -2], shape=(L, L - 2))
    w = np.ones(L)
    for i in range(niter):
        W = sparse.spdiags(w, 0, L, L)
        Z = W + lam * D.dot(D.T)
        z = spsolve(Z, w * y)
        w = p * (y > z) + (1 - p) * (y < z)
    return z


def correct_baseline(y, lam=1e5, p=0.1):
    baseline_values = baseline_als(y, lam=lam, p=p)
    return y - baseline_values


def single_pulse(y, method='max'):
    """
    Get amplitude of the single pulse

    Parameters
    ----------
    y : ndarray
        Oscillogram of the single pulse.

    method : str ('max', 'sum')

    """
    if method == 'sum':
        return np.sum(y[y >= 0], dtype=float)
    elif method == 'max':
        return max(y)


def random_pulse(y, method='max'):
    if method == 'sum':
        return np.sum(y, dtype=float)
    elif method == 'max':
        peaks, _ = find_peaks(y[y > 0], distance=50, width=10)
        return sum(y[y > 0][peaks])


def periodic_pulse(data, frequency, time_window, method='max'):
    discretedata = []
    points_period = int(1 / frequency / data.horizInterval) + 1
    points_window = int(time_window / data.horizInterval) + 1
    y = data.y
    init_point = np.argmax(y)
    pulses_points = np.append(
        np.arange(init_point, 0, -points_period)[::-1],
        np.arange(init_point, len(y), points_period))
    for p in pulses_points:
        if p < points_window:
            low = 0
            top = points_window
        else:
            low = p - points_window // 2
            top = p + points_window // 2
        discretedata.append(single_pulse(y[low:top], method=method))
    return discretedata


def scope_unwindowed(data, time_discrete, method='max'):
    points_discrete = int(time_discrete // data.horizInterval)
    y = data.y
    points_discrete += 1
    discretedata = [random_pulse(y[i:i + points_discrete], method=method)
                    for i in range(0, len(y), points_discrete)]
    return discretedata


def memo_oscillogram(data, vendor, correct_bs=True):
    if type(data) is str:
        filedata = (data, parse_file(data, vendor))
    if type(data) == tuple:
        if type(data[1]) is not str:
            return data

        filedata = (data[0], parse_file(data[0], vendor))
        filedata[1].y = data[1]
        return filedata

    y = filedata[1].y
    y -= np.min(y)
    filedata[1].y = correct_baseline(y) if correct_bs else y

    return filedata


@dataclass
class PulsesHistMaker:
    """
    Class for extract amplitude histogram from plenty of oscillogram files
    Supported formats are:
        - LecCroy binary .trc
        - Tektronix binary v.2 .wfm

    __init__ arguments
    ------------------

    datadir: str
        Path to oscillogram files
    fsnum: int = -1
        Number of files to process.
        If -1 all files in the datadir will be processed
    fsoffset: int = 0
        Skip files from the first ine in the datadir
    fchunksize: int = 10
        The size of chunk for parallel processing
    parallel: bool = False
        Use joblib or not?
    parallel_jobs: int = -1
        Number of parallel jobs
    memo_file: str = ''
        Path to file with compressed data from previous processing.
        Can be use iff use save it previously with 'save_memo' method
    histbins: int = 2000
        Numbere of bins in the histogram.
        Can be change in methods 'make_hist' and 'get_hist'
    correct_baseline: bool = True
        Correct baseline of not?
    method : str = 'max'
        Method of pulses counting. Can be 'max' or 'sum'
    vendor: str = 'lecroy'
        Vendor name. Can be 'tek' or 'lecroy'
    """

    datadir: str
    fsnum: int = -1
    fsoffset: int = 0
    fchunksize: int = 10
    parallel: bool = False
    parallel_jobs: int = -1
    memo_file: str = ''
    histbins: int = 2000
    correct_baseline: bool = True

    method: str = 'max'
    methods: tuple = ('sum', 'max')

    vendor: str = 'lecroy'
    vendors: tuple = ('lecroy', 'tek')

    def __post_init__(self):
        if not self.parallel:
            self.parallel_jobs = 1
        if self.vendor not in self.vendors:
            raise ValueError('vendor must be in %s, not %s' %
                             (self.vendors, self.vendor))
        if self.method not in self.methods:
            raise ValueError('method must be in %s, not %s' %
                             (self.methods, self.method))

    def read(self, fsnum=-1, parallel_read=False):
        if fsnum == -1:
            fsnum = self.fsnum
        self.rawdata = list_files(
            self.datadir, self.vendor, self.fsoffset, fsnum)
        if self.memo_file:
            with open(self.memo_file, 'rb') as f:
                memodata = load(f, compression='lzma',
                                set_default_extension=False)
            for k, path in enumerate(self.rawdata):
                if path in self.rawdata:
                    self.rawdata[k] = (path, memodata[path])

        self.filesnum = len(self.rawdata)

    def save_memo(self, filename):
        self.clear_rawdata()
        with open(filename, 'wb') as f:
            dump(dict(self.rawdata), f, compression='lzma',
                 set_default_extension=False)

    def save_hist(self, filename):
        np.savetxt(filename, np.vstack((self.bins[:-1], self.hist)).T)

    def clear_rawdata(self):
        for k in range(len(self.rawdata)):
            p, d = self.rawdata[k]
            self.rawdata[k] = (p, d.y)

    def single_pulse_hist(self, div_start=5.9, div_width=0.27):
        self.discretedata = []
        i = 1
        for d in self.rawdata:
            x, y = windowed(d, div_start, div_width)
            self.discretedata.append(single_pulse(y, self.method))
            i += 1
        self.make_hist(self.histbins)

    def multi_pulse_histogram(self, frequency=2.5e6, time_window=7.5e-9):
        self.parse(periodic_pulse, (frequency, time_window, self.method))
        self.make_hist(self.histbins)

    def unwindowed_histogram(self, time_discrete=15e-9):
        self.parse(scope_unwindowed, (time_discrete, self.method))
        self.make_hist(self.histbins)

    def parse(self, func, args):
        self.discretedata = []

        for i in range(0, self.filesnum, self.fchunksize):
            t = time.time()
            hb = min(i + self.fchunksize, self.filesnum)
            self.rawdata[i:hb] = Parallel(
                n_jobs=self.parallel_jobs)(
                    delayed(memo_oscillogram)(
                        df, self.vendor, self.correct_baseline)
                for df in self.rawdata[i:hb])
            pulsesdata = [func(df[1], *args) for df in self.rawdata[i:hb]]
            self.discretedata += pulsesdata

            print('Files ##%d-%d time %.2f s' %
                  (i, hb, time.time() - t), end='\t')

            del pulsesdata
            gc.collect()

    def make_hist(self, histbins):
        self.hist, self.bins = np.histogram(self.discretedata, bins=histbins)

    def get_hist(self, histbins):
        self.make_hist(histbins)
        return self.bins[:-1], self.hist



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  2 13:11:13 2020

@author: Pavel Gostev

numpy version
"""

import functools
import numpy as np
from scipy.special import factorial

__all__ = ['compose', 'lrange', 'fact', 'p_convolve', 'moment', 'mean', 'g2',
           'normalize', 'abssum', 'fidelity', 'entropy', 'DPREC']

try:
    DPREC = np.float128
except AttributeError:
    DPREC = np.float64
    print('%s:%s:' % (__file__, 19), 'RuntimeWarning:',
          "Numpy.float128 can not be used. DPREC is numpy.float64, results may be unprecise.")


def compose(*functions):
    def pack(x): return x if type(x) is tuple else (x,)

    return functools.reduce(
        lambda acc, f: lambda *y: f(*pack(acc(*pack(y)))), reversed(functions), lambda *x: x)


def lrange(iterable):
    return compose(np.arange, len)(iterable)


def fact(n):
    return DPREC(factorial(n, exact=True))


def p_convolve(p_signal, p_noise):
    N = max(len(p_signal), len(p_noise))
    return normalize(np.convolve(p_signal, p_noise)[:N])


def moment(p, N):
    return sum(i ** N * p[i] for i in lrange(p))


def mean(p):
    return moment(p, 1)


def g2(p):
    m = mean(p)
    s = moment(p, 2)
    if m == 0:
        m = 1
        print('Uncorrect g2. Zero mean change to 1')
    return (s - m) / m ** 2


def normalize(p):
    return np.array(p) / moment(p, 0)


def abssum(p):
    return sum((p[i]) ** 2 for i in lrange(p))


def fidelity(p1, p2):
    if len(p1) != len(p2):
        np.warnings.warn_explicit(
            'Probabilities for the fidelity calculation must have the same length',
            RuntimeWarning, __file__, 72)
    plen = min(len(p1), len(p2))
    prod = [p1[i] * p2[i] for i in range(plen)]
    return sum(np.sign(p)*np.sqrt(abs(p)) for p in prod) ** 2


def entropy(p):
    return sum(- e * np.log(e) for e in p if e > 0)



# -*- coding: utf-8 -*-
"""
Created on Tue Aug 27 20:51:49 2019

@author: Pavel Gostev
"""
import numpy as np
from functools import lru_cache

from scipy.optimize import brute, OptimizeResult
from scipy.stats import poisson
from . import (g2, mean, normalize, P2Q, moment)
import logging


log = logging.getLogger('crosstalk')
log.setLevel(logging.INFO)
info = log.info


def d_crosstalk_4n(p_crosstalk: float):
    """
    Probabilities of k ≤ 5 triggered pixels for the 4-neighbours model

    See table 1 in [1]

    Parameters
    ----------
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    ctnoise : ndarray
        Probability distributions of the total number
        of triggered pixels for a single initially fired pixel.

    References
    ----------
    .. [1]
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    """

    q_crosstalk = (1 - p_crosstalk)
    ctnoise = np.zeros(6)
    ctnoise[1] = q_crosstalk ** 4
    ctnoise[2] = 4 * p_crosstalk * q_crosstalk ** 6
    ctnoise[3] = 18 * p_crosstalk ** 2 * q_crosstalk ** 8
    ctnoise[4] = 4 * p_crosstalk ** 3 * q_crosstalk ** 8 * (
        1 + 3 * q_crosstalk + 18 * q_crosstalk ** 2)
    ctnoise[5] = 5 * p_crosstalk ** 4 * q_crosstalk ** 10 * (
        8 + 24 * q_crosstalk + 55 * q_crosstalk ** 2)
    return ctnoise


@lru_cache(maxsize=None)
def p_crosstalk_m(m: int, k: int, p_crosstalk: float):
    """
    The probability of total number k of triggered pixels provided m primaries

    See formula 2.20 in [1]

    Parameters
    ----------
    m : int
        Number of primary pixels.
    k : int
        Number of triggered pixels.
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    P_{m+1} : float
        The probability of total number k of triggered pixels
        provided m primaries


    References
    ----------
    .. [1]
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    """

    ctnoise = d_crosstalk_4n(p_crosstalk)
    if m == 1:
        if k > 5:
            return 0
        return ctnoise[k]
    return sum(p_crosstalk_m(m - 1, k - i, p_crosstalk) * ctnoise[i]
               for i in range(1, k - m + 2, 1) if i <= 5)


def distort(Qcorr: np.array, p_crosstalk: float):
    """
    Include crosstalk noise into photocounting statistics.
    We use model with 4 neighbors with saturation.

    See formula 2.19 in [1]

    Parameters
    ----------
    Qcorr : iterable
        The photocounting statistics.
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    Q : ndarray
        Noised photocounting statistics.

    References
    ----------
    .. [1]
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    """

    N = len(Qcorr)
    Q = np.zeros(N)
    Q[0] = Qcorr[0]

    @np.vectorize
    def point(k):
        return sum(Qcorr[m] * p_crosstalk_m(m, k, p_crosstalk) for m in range(1, k + 1, 1))

    Q[1:] = point(np.arange(1, N))
    return Q


def compensate(Q: np.array, p_crosstalk: float):
    """
    Remove crosstalk noise from photocounting statistics.
    We use model with 4 neighbors with saturation.

    See formula 2.21 in [1]


    Parameters
    ----------
    Q : iterable
        The photocounting statistics.
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    Qcorr : ndarray
        Denoised photocounting statistics.

    References
    ----------
    .. [1]
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    """
    N = len(Q)
    eps = total_pcrosstalk(p_crosstalk)
    Qcorr = np.zeros(N)
    Qcorr[0] = Q[0]
    Qcorr[1] = Q[1] / (1 - eps)
    for m in range(2, N):
        c1 = (1 - eps) ** -m
        c2 = Q[m] - sum(Qcorr[k] * p_crosstalk_m(k, m, p_crosstalk)
                        for k in range(1, m))
        Qcorr[m] = c1 * c2

    return normalize(Qcorr)


def optctp(_pct_param, Q, PDE, N, mtype, n_cells):
    """
    Optimization function

    Parameters
    ----------
    _pct_param : TYPE
        DESCRIPTION.
    Q : iterable
        Experimental photocounting statistics of a laser source.
    PDE : float
        PDE of the detector.
    N : int
        Size of poisson photon-number statistics.
    mtype : {'binomial', 'subbinomial'}, optional
        Type of the detector: ideal is binomial, realistic is subbinomial,
        but in the most of applications one can consider the detector as binomial
        The default is 'binomial'.
    n_cells : TYPE, optional
        Number of photocounting cells in the subbinomial case. The default is 0.

    Returns
    -------
    discrepancy : float
        Abcolute difference of g2 of poisson photocounting statistics
        and g2 of compensated experimental statistics.

    """

    p_crosstalk, poisson_mean = _pct_param
    P = poisson.pmf(np.arange(N), poisson_mean)
    Qtheory = P2Q(P, PDE, len(Q), mtype, n_cells)
    Qest = compensate(Q, p_crosstalk)
    return abs(g2(Qest) - g2(Qtheory))


def find_pcrosstalk(Q: np.array, PDE: float, N: int,
                    mtype: str = 'binomial', n_cells: int = 0,
                    Ns: int = 100, min_pct: float = 0, max_pct: float = 0.1):
    """
    Brute searching of crosstalk probability
    by optimizing of g2 difference from noised data and
    noised poisson photocounting statistics.
    We use model with 4 neighbors with saturation.

    See [1]

    Parameters
    ----------
    Q : iterable
        Experimental photocounting statistics of a laser source.
    PDE : float
        PDE of the detector.
    N : int
        Size of poisson photon-number statistics.
    mtype : {'binomial', 'subbinomial'}, optional
        Type of the detector: ideal is binomial, realistic is subbinomial,
        but in the most of applications one can consider the detector as binomial
        The default is 'binomial'.
    n_cells : TYPE, optional
        Number of photocounting cells in the subbinomial case. The default is 0.
    Ns : int, optional
        Size of calculation grid of brute function.
        The default is 100.

    Returns
    -------
    p_ct: float
        The probability of a single crosstalk event.

    res: OptimizeResult
        See scipy.optimize.brute description.
        res.x0 consists of optimal p_crosstalk and
        optimal poisson photon-number distribution
        p_crosstalk is in res.x0[0]

    References
    ----------
    .. [1]
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    """

    res = brute(optctp, ([min_pct, max_pct], [mean(Q) / PDE * 0.9,
                                              mean(Q) / PDE * 1.1]),
                args=(Q, PDE, N, mtype, n_cells), Ns=Ns, full_output=True,
                workers=-1)
    info("P_ct = {r[0][0]}, Δg(2) = {r[1]}".format(r=res))
    return res[0][0], OptimizeResult(x=res[0], fval=res[1],
                                     grid=res[2], Jout=res[3])


def Q2total_pcrosstalk(Q: np.array):
    """
    Calculate model independent total crosstalk probability.
    It's may be vary from the result of optimize_pcrosstalk

    See formula 2.23 in [1]

    Parameters
    ----------
    Q : iterable
        Experimental photocounting statistics of a pulsed laser source.

    Returns
    -------
    epsilon : float
        The total crosstalk probability.

    References
    ----------
    .. [1]
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    """

    mu = - np.log(Q[0])
    return 1 - Q[1] / mu / np.exp(- mu)


def total_pcrosstalk(p_crosstalk: float):
    """
    Calculate total crosstalk probability from
    the probability of single crosstalk event.
    We use model with 4 neighbors with saturation

    See text after formula 2.1 in [1]

    Parameters
    ----------
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    epsilon : float
        The total crosstalk probability.

    References
    ----------
    .. [1]
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    """

    return 1 - (1 - p_crosstalk)**4


def single_pcrosstalk(total_pcrosstalk: float):
    """
    Inverse formula to ``total_pcrosstalk''

    Parameters
    ----------
    total_pcrosstalk : float
        The probability of total crosstalk events.

    Returns
    -------
    p_crosstalk : float
        The probability of a single crosstalk event.

    """
    return 1 - (1 - total_pcrosstalk) ** 0.25


def ENF(p_crosstalk: float):
    """
    Calculate excess noise factor (ENF) of the detector

    See formulas 2.16, 2.17 and 2.27 in [1]

    Parameters
    ----------
    p_crosstalk : float
        The probability of a single crosstalk event.

    Returns
    -------
    ENF : float
        Excess noise factor.

    References
    ----------
    .. [1]
    Gallego, L., et al. "Modeling crosstalk in silicon photomultipliers."
    Journal of instrumentation 8.05 (2013): P05010.
    https://iopscience.iop.org/article/10.1088/1748-0221/8/05/P05010/pdf

    """

    d = d_crosstalk_4n(p_crosstalk)
    r = d[5] / (1 - sum(d[1:5]))
    e1 = moment(d[:5], 1) + d[5] * (1 + 4*r) / r ** 2
    var1 = moment(d[:5], 2) + d[5] * (2 + 7*r + 16 * r ** 2) / r ** 3 - e1 ** 2
    return 1 + var1 / e1 ** 2
#



# -*- coding: utf-8 -*-
"""
@author: Pavel Gostev
"""
from ._numpy_core import normalize, lrange

import numpy as np
from scipy.signal import find_peaks
from scipy.special import eval_hermitenorm
from scipy.optimize import minimize
import matplotlib.pyplot as plt


def loadhist(path, **kwargs):
    for dl in [',', ' ', '\t']:
        try:
            bins, hist = np.loadtxt(path, delimiter=dl, unpack=True, **kwargs)
        except (ValueError, TypeError, FileNotFoundError, IOError) as E:
            np.warnings.warn_explicit(
                'Histogram parsing with delimiter' + r' "%s": %s' % (dl, E),
                IOError, __file__, 17)
            continue
        else:
            return bins, hist

    raise ValueError("Can't parse the histogram from the file %s" % path)


def gauss_hermite_poly(x, norm_factor, peak_pos, sigma, h3, h4):
    w = (x - peak_pos) / sigma
    return norm_factor * np.exp(- w ** 2 / 2) * (1 + h3*eval_hermitenorm(3, w) + h4*eval_hermitenorm(4, w))


def peak_area(norm_factor, peak_pos, sigma, h3, h4):
    return norm_factor * sigma * (np.sqrt(2*np.pi) + h4)


def minpoly(popt, bins, hist):
    return np.linalg.norm(gauss_hermite_poly(bins, *popt) - hist, 1)


def mindowns(popt, bins, hist, downs):
    return np.linalg.norm((hist - gauss_hermite_poly(bins, *popt))[downs])


def construct_q_sum(hist, peaks, downs):
    Q = []
    for i in lrange(downs)[:-1]:
        for p in peaks:
            dl = downs[i]
            dt = downs[i+1]
            if p < dl or p > dt:
                continue
            Q.append(sum(hist[dl:dt]))
            break
    return Q


def construct_q_fit(hist, bins, peaks, downs):
    Q = []
    for i in lrange(downs)[:-1]:
        for p in peaks:

            dl = downs[i]
            dt = downs[i+1]

            if p < dl or p > dt:
                continue
            res = minimize(minpoly, args=(bins[dl:dt], hist[dl:dt]),
                           x0=(hist[p], bins[p], bins[1] -
                               bins[0], 0.0, 0.0),
                           bounds=list(zip(
                               [hist[p] / 2, bins[dl], bins[1] -
                                   bins[0], -0.01, -0.01],
                               [hist[p], bins[dt], np.sqrt(bins[dt] - bins[dl]) / 8, 0.01, 0.01])))
            popt = res.x
            Q.append(peak_area(*popt))
            # if plot:
            #     plt.plot(bins, gauss_hermite_poly(bins, *popt))
            break
    return Q


def hist2Q(hist, bins, discrete,
           threshold=1, peak_width=1, down_width=1,
           method='sum', remove_pedestal=1, lim_downssum=1, maxiter=20,
           plot=False, logplot=False):
    """
    Build photocounting statistics from an experimental histogram
    by gaussian-hermite polynoms or simple sum

    Parameters
    ----------
    hist : iterable
        The experimental histogram.
    discrete : int
        The amplitude of single photocount pulse in points.
    threshold : int, optional
        Minimal number of events to find histogram peak.
        The default is 1.
    peak_width : float, optional
        The width of peaks.
        It must be greater than 1 if the histogram is made by oscilloscope or 'max' method.
        The default is 1.
    down_width : float, optional
        The width of downs.
        It must be greater than 1 if the histogram is made by oscilloscope or 'max' method.
        The default is 1.
    plot : bool, optional
        Flag to plot hist and results of find_peaks.
        The default is False.
    logplot : bool
        Enable log yscale for histogram plotting.
        The default is False.        
    method : {'sum', 'fit', 'manual'}
        Method of the photocounting statistics construction.
            'sum' is a simple summation between minimums of the histogram

            'fit' is a gauss-hermite function fitteing like in [1]

            'manual' is a simple summation of intervals with fixed length
    remove_pedestal : boolean
        Flag to remove pedestal noise from the histogram
    lim_downssum : float
        Limit for sum of downs of the histogram
    maxiter : int
        Maximum iterations count for pedestal removing

    Returns
    -------
    Q : ndarray
        The photocounting statistics.

    References
    ----------
    .. [1]
    Ramilli, Marco, et al. "Photon-number statistics with silicon photomultipliers."
    JOSA B 27.5 (2010): 852-862.

    """
    discrete = int(discrete * 0.9)
    downs, _ = find_peaks(-hist, distance=discrete, width=down_width)

    if remove_pedestal:
        it = 0
        while sum(hist[downs]) > lim_downssum and it < maxiter:
            middle_point = bins[len(bins) // 2]
            dres = minimize(mindowns, args=(bins, hist, downs),
                            x0=(1, middle_point, np.sqrt(middle_point), 0, 0))
            hist -= gauss_hermite_poly(bins, *dres.x)
            hist[hist < 0] = 0
            downs, _ = find_peaks(-hist, distance=discrete, width=down_width)
            it += 1

        if plot:
            plt.plot(bins, hist)

    peaks, _ = find_peaks(np.concatenate(([0], hist)), threshold=threshold, distance=discrete,
                          width=peak_width, plateau_size=(0, 10))
    if peaks == []:
        raise ValueError('Histogram peaks were not found with given settings')
    peaks -= 1

    if plot:
        plt.scatter(bins[peaks], hist[peaks])
        plt.scatter(bins[downs], hist[downs])

    if method == 'manual':
        Q = []
        discrete = peaks[2] - peaks[1]
        for p in peaks:
            low = int(max(0, p - discrete // 2))
            top = int(min(p + discrete // 2, len(hist) - 1))
            Q.append(sum(hist[low:top]))

            if plot:
                plt.axvline(bins[low])
                plt.axvline(bins[top])

        if top != len(hist) - 1:
            Q.append(sum(hist[top:]))

    else:
        if method == 'sum':
            Q = construct_q_sum(hist, peaks, downs)
        elif method == 'fit':
            Q = construct_q_fit(hist, bins, peaks, downs)

    if plot:
        plt.plot(bins, hist)
        plt.xlabel('Amplitude, V')
        plt.ylabel("Events' number")
        plt.show()

    return normalize(Q)


class QStatisticsMaker:
    """
    Class to make photocounting statistics from histogram

    __init__ arguments
    ----------
    fname : string
        File name contains the histogram.
    photon_discrete : float
        The amplitude of the single-photocount pulse.
    peak_width : int, optional
        The width of peaks.
        It must be 1 if the histogram is made by 'count' method.
        It must be greater if the histogram is made by oscilloscope or 'max' method.
        The default is 1.
    down_width : float, optional
        The width of downs.
        It must be greater than 1 if the histogram is made by oscilloscope or 'max' method.
        The default is 1.
    plot : bool, optional
        Flag to plot hist and results of find_peaks.
        The default is False.
    logplot : bool
        Enable log yscale for histogram plotting.
        The default is False.        
    method : {'sum', 'fit', 'manual'}
        Method of the photocounting statistics construction.
            'sum' is a simple summation between minimums of the histogram

            'fit' is a gauss-hermite function fitteing like in [1]

            'manual' is a simple summation of intervals with fixed length
    skiprows : int, optional
        Number of preamble rows in the file. The default is 0.
    plot : bool, optional
        Flag to plot hist and results of find_peaks.
        The default is False.
    remove_pedestal : boolean
        Flag to remove pedestal from the histogram

    References
    ----------
    .. [1]
    Ramilli, Marco, et al. "Photon-number statistics with silicon photomultipliers."
    JOSA B 27.5 (2010): 852-862.

    """

    def __init__(self, fname, photon_discrete,
                 peak_width=1, down_width=1, method='fit', skiprows=0,
                 plot=False, logplot=False,
                 remove_pedestal=True):
        self.photon_discrete = photon_discrete
        self.fname = fname

        self._extract_data(skiprows)
        self.Q = hist2Q(self.hist, self.bins, discrete=self.points_discrete,
                        peak_width=peak_width, down_width=down_width,
                        plot=plot, logplot=logplot,
                        method=method,
                        remove_pedestal=remove_pedestal)

    # Reading information from the file
    def _extract_data(self, skiprows):
        self.bins, self.hist = loadhist(self.fname, skiprows=skiprows)
        if self.bins[1] - self.bins[0] < 0:
            self.bins = self.bins[::-1]
            self.hist = self.hist[::-1]
        self.points_discrete = int(self.photon_discrete /
                                   (self.bins[1] - self.bins[0]))

    def getq(self):
        """
        Returns the photocounting statistics was made

        Returns
        -------
        self.Q : ndarray
            self.Q[self.Q > 0].

        """

        return normalize(self.Q[self.Q > 0])



from ._numpy_core import *
from ._dcore import P2Q, Q2P
from .histogram import hist2Q, QStatisticsMaker
from .oscillogram import PulsesHistMaker
from .crosstalk import (find_pcrosstalk, compensate, distort,
                        Q2total_pcrosstalk, total_pcrosstalk, single_pcrosstalk,
                        ENF)

__version__ = "0.3.0"



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="oscsipm",
    version="0.3.0",
    author="Pavel Gostev",
    author_email="gostev.pavel@physics.msu.ru",
    description="Instruments to make photocounting statistics from histograms and raw oscillograms",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vongostev/OscSiPM",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    include_package_data=True,
    install_requires=[
        'numpy',
        'scipy',
        'matplotlib',
        'lecroyparser',
        'joblib',
        'dataclasses',
        'compress_pickle'
    ],
)



# -*- coding: utf-8 -*-
"""
Created on Mon Sep 16 23:10:39 2019

@author: vonGostev
"""
import numpy as np
from scipy.special import comb
from scipy.linalg import pinv
from functools import lru_cache

from ._numpy_core import DPREC


def d_binomial(qe: float, N: int, M: int):
    m = np.arange(M).reshape((-1, 1))
    n = np.arange(N).reshape((1, -1))
    d_matrix = comb(n, m, exact=False) * qe ** m * \
        (1 - qe)**(n - m)
    return d_matrix


def invd_binomial(qe: float, N: int, M: int, n_cells: int = 0):
    return d_binomial(1 / qe, N, M)


@lru_cache(maxsize=None)
def sb_elem(qe: float, n_cells: int, n: int, m: int):
    """
    Calculated from formula (5) via recursive approach from Appendix A

    Sperling, J., Vogel, W., & Agarwal, G. S. (2014).
    Quantum state engineering by click counting.
    Physical Review A, 89(4), 043829.
    for arbitrary quantum efficiency and zero noise

    Idea from
    Sperling, J., Vogel, W., & Agarwal, G. S. (2012).
    True photocounting statistics of multiple on-off detectors.
    Physical Review A, 85(2), 023820.

    """
    if n == 0 and m == 0:
        return 1
    elif n != 0 and m == 0:
        return (1 - qe) ** n
    elif m > n:
        return 0
    return (1 - qe + qe * m / n_cells) * sb_elem(qe, n_cells, n - 1, m) +\
        qe * (n_cells - m - 1) / n_cells * sb_elem(qe, n_cells, n - 1, m - 1)


def d_subbinomial(qe: float, N: int, M: int, n_cells: int = 1000):
    qe = DPREC(qe)
    m = np.arange(M).reshape((-1, 1))
    n = np.arange(N).reshape((1, -1))
    d_matrix = np.vectorize(sb_elem, otypes=[DPREC])(qe, n_cells, n, m)
    return d_matrix


def invd_subbinomial(qe: float, N: int, M: int, n_cells: int = 1000):
    return pinv(d_subbinomial(qe, N, M, n_cells))
